"""
monobit.storage - load and save fonts

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import sys
import logging
from pathlib import Path
from contextlib import contextmanager

from .constants import VERSION, CONVERTER_NAME
from .containers import ContainerFormatError, open_container
from .font import Font
from .pack import Pack
from .streams import MagicRegistry, FileFormatError, open_stream, maybe_text
from .scripting import scriptable, ScriptArgs, ARG_PREFIX
from .basetypes import Any


DEFAULT_TEXT_FORMAT = 'yaff'
DEFAULT_BINARY_FORMAT = 'raw'


##############################################################################

@contextmanager
def open_location(file, mode, where=None, overwrite=False):
    """
    Open a binary stream on a container or filesystem
    both `file` and `where` may be Streams, files, or file/directory names
    `where` may also be a Container
    if `where` is empty, the whole filesystem is taken as the container/location.
    if `overwrite` is True, will overwrite `file`. Note that `where` is always considered overwritable
    returns a Steam and a Container object
    """
    if mode not in ('r', 'w'):
        raise ValueError(f"Unsupported mode '{mode}'.")
    if not file and not where:
        raise ValueError(f'No location provided.')
    # interpret incomplete arguments
    # no choice - can't open a stream on a directory
    if isinstance(file, (str, Path)) and Path(file).is_dir():
        where = file
        file = None
    # only container location provided - traverse into it
    if where and not file:
        with open_container(where, mode, overwrite=True) as container:
            # empty file parameter means 'load/save all'
            yield None, container
        return
    if not where and isinstance(file, (str, Path)):
        # see if file is itself a container
        # don't open containers if we only have a stream - we don't want surprise directory creation
        try:
            with open_container(file, mode, overwrite=overwrite) as container:
                yield None, container
            return
        except ContainerFormatError as e:
            # file is not itself a container, use enclosing dir as container
            where = Path(file).parent
            file = Path(file).name
    # we have a stream and maybe a container
    with open_container(where, mode, overwrite=True) as container:
        with open_stream(file, mode, where=container, overwrite=overwrite) as stream:
            # see if file is itself a container
            try:
                with open_container(stream, mode, overwrite=overwrite) as container:
                    yield None, container
                return
            except ContainerFormatError as e:
                pass
            # infile is not a container, load/save single file
            yield stream, container


##############################################################################
# loading


@scriptable(unknown_args='passthrough', record=False)
def load(infile:Any='', *, format:str='', where:Any='', **kwargs):
    """
    Read font(s) from file.

    infile: input file (default: stdin)
    format: input format (default: infer from magic number or filename)
    where: enclosing container stream or name (default: current working directory)
    """
    infile = infile or sys.stdin
    return _load_from_location(infile, where, format, **kwargs)


def _load_from_location(infile, where, format, **kwargs):
    """Read font(s) from file."""
    # if container/file provided as string or steam, open them
    with open_location(infile, 'r', where=where) as (stream, container):
        # infile not provided - load all from container
        if not stream:
            return _load_all(container, format, **kwargs)
        return _load_from_file(stream, container, format, **kwargs)

def _load_from_file(instream, where, format, **kwargs):
    """Open file and load font(s) from it."""
    # identify file type
    loader = loaders.get_for(instream, format=format, do_open=True)
    if not loader:
        raise FileFormatError('Cannot load from format `{}`.'.format(format)) from None
    logging.info('Loading `%s` on `%s` as %s', instream.name, where.name, loader.name)
    fonts = loader(instream, where, **kwargs)
    # convert font or pack to pack
    if not fonts:
        raise FileFormatError('No fonts found in file.')
    pack = Pack(fonts)
    # set conversion properties
    filename = Path(instream.name).name
    # if the source filename contains surrogate-escaped non-utf8 bytes
    # preserve the byte values as backslash escapes
    try:
        filename.encode('utf-8')
    except UnicodeError:
        filename = (
            filename.encode('utf-8', 'surrogateescape')
            .decode('ascii', 'backslashreplace')
        )
    return Pack(
        _font.modify(
            converter=CONVERTER_NAME,
            source_format=_font.source_format or loader.name,
            source_name=_font.source_name or filename
        )
        for _font in pack
    )

def _load_all(container, format, **kwargs):
    """Open container and load all fonts found in it into one pack."""
    logging.info('Reading all from `%s`.', container.name)
    packs = Pack()
    # try opening a container on input file for read, will raise error if not container format
    for name in container:
        logging.debug('Trying `%s` on `%s`.', name, container.name)
        with open_stream(name, 'r', where=container) as stream:
            try:
                pack = _load_from_location(stream, where=container, format=format, **kwargs)
            except Exception as exc:
                # if one font fails for any reason, try the next
                # loaders raise ValueError if unable to parse
                logging.debug('Could not load `%s`: %s', name, exc)
            else:
                packs += Pack(pack)
    return packs


##############################################################################
# saving



@scriptable(unknown_args='passthrough', record=False, pack_operation=True)
def save(
        pack_or_font,
        outfile:Any='', *,
        format:str='', where:Any='', overwrite:bool=False,
        **kwargs
    ):
    """
    Write font(s) to file.

    outfile: output file (default: stdout)
    format: font file format
    where: enclosing location/container. (default: current working directory)
    overwrite: if outfile is a filename, allow overwriting existing file
    """
    # `where` is mandatory for formats that need filesystem access.
    # if specified and outfile is a filename, it is taken relative to this location.
    pack = Pack(pack_or_font)
    outfile = outfile or sys.stdout
    if outfile == sys.stdout:
        # errors can occur if the strings we write contain surrogates
        # these may come from filesystem names using 'surrogateescape'
        sys.stdout.reconfigure(errors='replace')
    with open_location(outfile, 'w', where=where, overwrite=overwrite) as (stream, container):
        if not stream:
            _save_all(pack, container, format, **kwargs)
        else:
            _save_to_file(pack, stream, container, format, **kwargs)
    return pack_or_font

def _save_all(pack, where, format, **kwargs):
    """Save fonts to a container."""
    logging.info('Writing all to `%s`.', where.name)
    for font in pack:
        # generate unique filename
        name = font.name.replace(' ', '_')
        format = format or DEFAULT_TEXT_FORMAT
        filename = where.unused_name(name, format)
        try:
            with open_stream(filename, 'w', where=where) as stream:
                _save_to_file(Pack(font), stream, where, format, **kwargs)
        except BrokenPipeError:
            pass
        except Exception as e:
            logging.error('Could not save `%s`: %s', filename, e)
            #raise

def _save_to_file(pack, outfile, where, format, **kwargs):
    """Save fonts to a single file."""
    saver = savers.get_for(outfile, format=format, do_open=False)
    if not saver:
        raise FileFormatError('Cannot save to format `{}`.'.format(format))
    logging.info('Saving `%s` on `%s` as %s.', outfile.name, where.name, saver.name)
    saver(pack, outfile, where, **kwargs)


##############################################################################
# loader/saver registry

class ConverterRegistry(MagicRegistry):
    """Loader/Saver registry."""

    def __init__(self, func_name):
        """Set up registry and function name."""
        super().__init__()
        self._func_name = func_name

    def get_for_location(self, file, format='', where='', do_open=True):
        """Get loader/saver for font file location."""
        if not file and not where:
            return self.get_for(format=format)
        if where:
            # if container/file provided as string or steam, open them to check magic bytes
            with open_location(file, 'r', where=where) as (stream, container):
                # identify file type if possible
                return self.get_for(stream, format=format, do_open=do_open)
        else:
            return self.get_for(file, format=format, do_open=do_open)


    def get_for(self, file=None, format='', do_open=False):
        """
        Get loader/saver function for this format.
        infile must be a Stream or empty
        """
        converter = None
        if not format:
            converter = self.identify(file, do_open=do_open)
        if not converter:
            if format:
                converter = self[format]
            elif (
                    not file
                    or not file.name or file.name == '<stdout>'
                    or (do_open and maybe_text(file))
                ):
                converter = self[DEFAULT_TEXT_FORMAT]
            elif do_open:
                converter = self[DEFAULT_BINARY_FORMAT]
            else:
                if format:
                    msg = f'Format `{format}` not recognised'
                else:
                    msg = 'Could not determine format'
                    if file:
                        msg += f' from file name `{file.name}`'
                    msg += '. Please provide a -format option'
                raise ValueError(msg)
        return converter

    def get_args(self, file=None, format='', do_open=False):
        """
        Get loader/saver arguments for this format.
        infile must be a Stream or empty
        """
        converter = self.get_for(file, format, do_open)
        if not converter:
            return ScriptArgs()
        return converter.script_args

    def register(self, *formats, magic=(), name='', linked=None):
        """
        Decorator to register font loader/saver.

        *formats: extensions covered by registered function
        magic: magic sequences covered by the converter (no effect for savers)
        name: name of the format
        linked: loader/saver linked to saver/loader
        """
        register_magic = super().register

        def _decorator(original_func):
            # set script arguments
            funcname = self._func_name
            if name:
                funcname += f' {ARG_PREFIX}format={name}'
            _func = scriptable(
                original_func,
                # use the standard name, not that of the registered function
                name=funcname,
                # don't record history of loading from default format
                record=(DEFAULT_TEXT_FORMAT not in formats),
            )
            # register converter
            if linked:
                linked.linked = _func
                _func.name = name or linked.name
                _func.formats = formats or linked.formats
                _func.magic = magic or linked.magic
            else:
                _func.name = name
                _func.linked = linked
                _func.formats = formats
                _func.magic = magic
            # register magic sequences
            register_magic(_func.name, *_func.formats, magic=_func.magic)(_func)
            return _func

        return _decorator


loaders = ConverterRegistry('load')
savers = ConverterRegistry('save')



"""
monobit.formats.mac - Mac OS fonts

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from ...storage import loaders, savers

from .dfont import _parse_mac_resource
from .nfnt import _extract_nfnt, _convert_nfnt
from .lisa import _load_lisa


# the magic is optional - a 'maybe magic'
# .rsrc is what we use as a 'filename' for resources inside containers
@loaders.register('dfont', 'suit', 'rsrc', name='mac', magic=(b'\0\0\1\0\0',))
def load_mac_dfont(instream, where=None):
    """Load font from a MacOS suitcase."""
    data = instream.read()
    return _parse_mac_resource(data)

# \x90\0 is not a formal signature, but the most common set of FONT_TYPE flags
# the \x80 sigs are LISA compressed NFNTs
@loaders.register(
    'nfnt', 'f',
    name='nfnt',
    magic=(b'\x90\0', b'\xb0\0', b'\x90\x80', b'\xb0\x80')
)
def load_nfnt(instream, where=None, offset:int=0):
    """
    Load font from a bare FONT/NFNT resource.

    offset: starting offset in bytes of the NFNT record in the file (default 0)
    """
    instream.seek(offset)
    data = instream.read()
    fontdata = _extract_nfnt(data, 0)
    return _convert_nfnt({}, **fontdata)


@loaders.register(name='lisa')
def load_lisa(instream, where=None):
    """Load a LISA font library."""
    return _load_lisa(instream)



"""
monobit.streams - file stream tools

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import io
import sys
import logging
from pathlib import Path


# number of bytes to read to check if something looks like text
_TEXT_SAMPLE_SIZE = 256
# bytes not expected in (modern) text files
_NON_TEXT_BYTES = (
    # C0 controls except HT, LF, CR
    tuple(range(9)) + (11, 12,) + tuple(range(14, 32))
    # also check for F8-FF which shouldn't occur in utf-8 text
    + tuple(range(0xf8, 0x100))
    # we don't currently parse text formats that need the latin-1 range:
    # - yaff is utf-8 excluding controls
    # - bdf, bmfont are printable ascii [0x20--0x7e] plus 0x0a, 0x0d
    # - hex, draw have undefined range, but we can assume ascii or utf-8
)


class FileFormatError(Exception):
    """Incorrect file format."""


def open_stream(file, mode, *, where=None, overwrite=False):
    """Ensure file is a stream of the right type, open or wrap if necessary."""
    return Stream(file, mode, where=where, overwrite=overwrite)


class StreamBase:
    """Shared base for stream and container."""

    def __init__(self, stream, mode='', name=''):
        self._stream = stream
        self.name = name
        self.mode = mode[:1]
        if self._stream:
            if not self.name:
                self.name = get_name(stream)
            if not self.mode:
                if self._stream.readable():
                    self.mode = 'r'
                else:
                    self.mode = 'w'
        self._refcount = 0
        self.closed = False

    def __enter__(self):
        self._refcount += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure archive is closed and essential records written."""
        if exc_type == BrokenPipeError:
            return True
        self._refcount -= 1
        logging.debug('Exiting %r with reference count %d.', self, self._refcount)
        if not self._refcount:
            self.close()

    def __repr__(self):
        """String representation."""
        return (
            f"<{type(self).__name__} name='{self.name}' mode='{self.mode}'"
            f"{' [closed]' if self.closed else ''}>"
        )

    def close(self):
        if self._stream and not self._refcount:
            self._stream.close()
        self.closed = True


class StreamWrapper(StreamBase):
    """Wrapper object to emulate a single text stream."""

    def __getattr__(self, attr):
        """Delegate undefined attributes to wrapped stream."""
        return getattr(self._stream, attr)

    def __iter__(self):
        # dunder methods not delegated
        return self._stream.__iter__()


class KeepOpen(StreamWrapper):
    """Wrapper to avoid closing wrapped stream."""

    def close(self):
        """Don't close underlying stream."""
        self._stream.flush()


class Stream(StreamWrapper):
    """Manage file resource."""

    def __init__(self, file, mode, *, name='', where=None, overwrite=False):
        """
        Ensure file is a stream of the right type, open or wrap if necessary.
            file: stream, string or path-like object
            mode: 'r' or 'w'
            where: container to open any new stream on
        """
        if not file:
            raise ValueError('No file name, path or stream provided.')
        mode = mode[:1]
        # if a path is provided, open a (binary) stream
        if isinstance(file, (str, Path)):
            file = self._open_path(file, mode, where, overwrite)
            self._raw = file
        else:
            # don't close externally provided stream
            file = KeepOpen(file)
            self._raw = None
        # initialise wrapper
        super().__init__(file, mode=mode, name=name)
        # check r/w mode is consistent
        self._ensure_rw()
        # placeholder for text wrapper
        self._textstream = None
        # Ensure we have a binary stream
        self._ensure_binary()

    @staticmethod
    def _open_path(file, mode, where, overwrite):
        """Open a raw stream on path and container provided."""
        path = Path(file)
        if not where:
            # open on filesystem
            if not overwrite and mode == 'w' and path.exists():
                raise FileExistsError(
                    f'Use option `-overwrite` to replace existing file `{file}`.'
                )
            logging.debug("Opening file `%s` for mode '%s'.", file, mode)
            file = io.open(path, mode + 'b')
        else:
            # open on container
            if not overwrite and mode == 'w' and file in where:
                raise FileExistsError(
                    f'Use option `-overwrite` to replace existing file `{file}` on `{where.name}`.'
                )
            file = where.open(path, mode)
        return file

    def _ensure_rw(self):
        """Ensure r/w mode is consistent."""
        if self.mode == 'r' and not self._stream.readable():
            raise FileFormatError('Expected readable stream, got writable.')
        if self.mode == 'w' and not self._stream.writable():
            raise FileFormatError('Expected writable stream, got readable.')

    def _ensure_binary(self):
        """Ensure we have a binary stream."""
        # a text format can be read from/written to a binary stream with a wrapper
        if not is_binary(self._stream):
            self._textstream = self._stream
            try:
                self._stream = self._stream.buffer
            except AttributeError as e:
                raise FileFormatError('Unable to access binary stream.') from e
            logging.debug('Getting buffer %r from text stream %r.', self._stream, self._textstream)

    @property
    def text(self):
        """Return underlying text stream or wrap underlying binary stream with utf-8 wrapper."""
        if not self._textstream:
            encoding = 'utf-8-sig' if self.mode == 'r' else 'utf-8'
            self._textstream = io.TextIOWrapper(
                self._stream, encoding=encoding,
                # on the one hand, this avoids breaks on slightly damaged files
                # on the other hand, we are less likely to break
                # on files that are clearly not text
                errors='ignore'
            )
        return self._textstream

    def __getattr__(self, attr):
        """Delegate undefined attributes to wrapped stream."""
        return getattr(self._stream, attr)

    def close(self):
        """Close stream, absorb errors."""
        # always close at wrapper level
        self.closed = True
        if self._textstream:
            try:
                self._textstream.close()
            except EnvironmentError:
                pass
        try:
            super().close()
        except EnvironmentError:
            pass
        if self._raw:
            try:
                self._raw.close()
            except EnvironmentError:
                pass


def is_binary(stream):
    """Check if stream is binary."""
    if stream.readable():
        # read 0 bytes - the return type will tell us if this is a text or binary stream
        return isinstance(stream.read(0), bytes)
    # write empty bytes - error if text stream
    try:
        stream.write(b'')
    except TypeError:
        return False
    return True

def get_name(stream):
    """Get stream name, if available."""
    try:
        return stream.name
    except AttributeError:
        # not all streams have one (e.g. BytesIO)
        return ''


###################################################################################################
# recognise file types

def normalise_suffix(suffix):
    """Bring suffix to lowercase without dot."""
    if suffix.startswith('.'):
        suffix = suffix[1:]
    return suffix.lower()

def get_suffix(file):
    """Get normalised suffix for file or path."""
    if isinstance(file, (str, Path)):
        suffix = Path(file).suffix
    else:
        suffix = Path(get_name(file)).suffix
    return normalise_suffix(suffix)

def has_magic(instream, magic):
    """Check if a binary stream matches the given signature."""
    try:
        return instream.peek(len(magic)).startswith(magic)
    except EnvironmentError:
        # e.g. write-only stream
        return False

def maybe_text(instream):
    """
    Check if a binary input stream looks a bit like it might hold utf-8 text.
    Currently just checks for unexpected bytes in a short sample.
    """
    if instream.writable():
        # output binary streams *could* hold text
        # (this is not about the file type, but about the content)
        return True
    try:
        sample = instream.peek(_TEXT_SAMPLE_SIZE)
    except EnvironmentError:
        return None
    if set(sample) & set(_NON_TEXT_BYTES):
        logging.debug(
            'Found unexpected bytes: identifying unknown input stream as binary.'
        )
        return False
    try:
        sample.decode('utf-8')
    except UnicodeDecodeError as err:
        # need to ensure we ignore errors due to clipping inside a utf-8 sequence
        if err.reason != 'unexpected end of data':
            logging.debug(
                'Found non-UTF8: identifying unknown input stream as binary.'
            )
            return False
    logging.debug('Tentatively identifying unknown input stream as text.')
    return True


class MagicRegistry:
    """Registry of file types and their magic sequences."""

    def __init__(self):
        """Set up registry."""
        self._magic = {}
        self._suffixes = {}

    def register(self, *suffixes, magic=()):
        """Decorator to register class that handles file type."""
        def decorator(klass):
            for suffix in suffixes:
                suffix = normalise_suffix(suffix)
                self._suffixes[suffix] = klass
            for sequence in magic:
                self._magic[sequence] = klass
            # sort the magic registry long to short to manage conflicts
            self._magic = {
                _k: _v for _k, _v in sorted(
                    self._magic.items(),
                    key=lambda _i:len(_i[0]), reverse=True
                )
            }
            # use first suffix given as standard
            if suffixes:
                klass.format = normalise_suffix(suffixes[0])
            return klass
        return decorator

    def __contains__(self, suffix):
        """Suffix is covered."""
        return normalise_suffix(suffix) in self._suffixes.keys()

    def __getitem__(self, suffix):
        """Get type by suffix."""
        return self._suffixes.get(suffix, None)

    def identify(self, file, do_open=False):
        """Identify a type from magic sequence on input file."""
        if not file:
            return None
        # can't read magic on write-only file
        if do_open:
            if isinstance(file, (str, Path)):
                # only use context manager if string provided
                # if we got an open stream we should not close it
                with open_stream(file, 'r') as stream:
                    return self.identify(stream, do_open=do_open)
            for magic, klass in self._magic.items():
                if has_magic(file, magic):
                    logging.debug(
                        'Magic bytes %a: identifying stream as %s.',
                        magic.decode('latin-1'), klass.__name__
                    )
                    return klass
        suffix = get_suffix(file)
        converter = self[suffix]
        if converter:
            logging.debug(
                'Filename suffix `%s`: identifying stream as %s.',
                suffix, converter.__name__
            )
        return converter



"""
monobit.formats.geos - C64 GEOS font files

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import count

from ..storage import loaders, savers
from ..font import Font
from ..glyph import Glyph
from ..raster import Raster
from ..struct import little_endian as le
from ..binary import ceildiv, align
from .raw import load_binary


# offset CVT signature b'formatted GEOS file' could be used as magic
@loaders.register('cvt', name='geos')
def load_geos(instream, where=None,  merge_mega:bool=True):
    """Load fonts from a GEOS ConVerT container."""
    return _load_geos_cvt(instream, merge_mega)

@loaders.register('vlir', name='geos-vlir')
def load_geos_vlir(instream, where=None, offset:int=0):
    """
    Load a bare GEOS font VLIR.

    offset: starting offset in bytes of the VLIR record in the file (default 0)
    """
    instream.seek(offset)
    return _load_geos_vlir(instream)


###############################################################################
# GEOS font VLIR
# https://www.lyonlabs.org/commodore/onrequest/geos/geos-fonts.html

_HEADER = le.Struct(
    baseline='uint8',
    stride='uint16',
    height='uint8',
    index_offset='uint16',
    bitstream_offset='uint16',
)

# characters 0x20 - 0x7f
# do we get plus one for the offset to the end?
_OFFSETS = le.uint16.array(96)


def _load_geos_vlir(instream):
    """Load a bare GEOS font VLIR."""
    anchor = instream.tell()
    header = _HEADER.read_from(instream)
    logging.debug(header)
    instream.seek(anchor + header.index_offset)
    offsets = _OFFSETS.read_from(instream)
    instream.seek(anchor + header.bitstream_offset)
    strikebytes = instream.read(header.height * header.stride)
    strikebytes = strikebytes.ljust(header.height * header.stride, b'\0')
    strike = Raster.from_bytes(
        strikebytes,
        header.stride * 8, header.height,
    )
    # clip out glyphs
    glyphs = tuple(
        Glyph(
            strike.crop(left=_offset, right=max(0, header.stride*8 - _next)),
            codepoint=_cp,
            shift_up=-(header.height-header.baseline-1)
        )
        for _offset, _next, _cp in zip(offsets, offsets[1:], count(0x20))
    )
    props = dict(
        descent=header.height-header.baseline-1,
        ascent=header.baseline+1,
    )
    return Font(glyphs, **props)


###############################################################################
# CVT container
# https://ist.uwaterloo.ca/~schepers/formats/GEOS.TXT
# https://ist.uwaterloo.ca/~schepers/formats/CVT.TXT

# the CVT contains a signature block, info block and record block, followed by
# one or more VLIRs (whose starting sectors are given by the record block)
# each such record is one sectore, i.e. 256 bytes.
# However in the CVT files the two initial bytes, a linked-list pointer,
# are left out.

# https://ist.uwaterloo.ca/~schepers/formats/D64.TXT
# > File Type
# > $00 - Scratched (deleted file entry)
# >  80 - DEL
# >  81 - SEQ
# >  82 - PRG
# >  83 - USR
# >  84 - REL
_C64_FILETYPES = {
    0x00: 'Scratched',
    0x80: 'DEL',
    0x81: 'SEQ',
    0x82: 'PRG',
    0x83: 'USR',
    0x84: 'REL',
}

# > GEOS file structure
# >   $00 - Sequential
# >    01 - VLIR file
_GEOS_STRUCTURES = {
    0x00: 'Sequential',
    0x01: 'VLIR',
}
# > GEOS filetype
# >    $00 - Non-GEOS (normal C64 file)
# >     01 - BASIC
# >     02 - Assembler
# >     03 - Data file
# >     04 - System File
# >     05 - Desk Accessory
# >     06 - Application
# >     07 - Application Data (user-created documents)
# >     08 - Font File
# >     09 - Printer Driver
# >     0A - Input Driver
# >     0B - Disk Driver (or Disk Device)
# >     0C - System Boot File
# >     0D - Temporary
# >     0E - Auto-Execute File
# >  0F-FF - Undefined
_GEOS_FILETYPES = {
    0x00: 'Non-GEOS',
    0x01: 'BASIC',
    0x02: 'Assembler',
    0x03: 'Data file',
    0x04: 'System File',
    0x05: 'Desk Accessory',
    0x06: 'Application',
    0x07: 'Application Data (user-created documents)',
    0x08: 'Font File',
    0x09: 'Printer Driver',
    0x0A: 'Input Driver',
    0x0B: 'Disk Driver (or Disk Device)',
    0x0C: 'System Boot File',
    0x0D: 'Temporary',
    0x0E: 'Auto-Execute File',
}

_GEOS_FONT_TYPE = 0x08

# 30 bytes 0x1e
# https://ist.uwaterloo.ca/~schepers/formats/GEOS.TXT
_DIR_BLOCK = le.Struct(
    # > C64 filetype (see the section on D64 for an explanation)
    # > REL files are not allowed.
    filetype='uint8',
    # > Starting track/sector (02/02 from above) of C64 file if GEOS
    # > filetype is $00. If GEOS filetype is non-zero,  track/sector
    # > of single-sector RECORD block
    sector='uint16',
    # > Filename (in ASCII, padded with $A0, case varies)
    filename='16s',
    # > Track/sector location of info block
    info_sector='uint16',
    # > GEOS file structure
    geos_structure='uint8',
    # > GEOS filetype
    geos_filetype='uint8',
    # > Year (1900 + value)
    year='uint8',
    # > Month (1-12, $01 to $0C)
    month='uint8',
    # > Day (1-31, $01 to $1F)
    day='uint8',
    # > Hour (0-23, $00 to $17) in military format
    hour='uint8',
    # > Minute (0-59, $00 to $3B)
    minute='uint8',
    # > Filesize, in sectors (low/high byte order)
    filesize='uint16',
)

# https://ist.uwaterloo.ca/~schepers/formats/CVT.TXT
_SIG_BLOCK = le.Struct(
    # 0x1e   30
    # b'PRG formatted GEOS file V1.0'
    # b'SEQ formatted GEOS file V1.0'
    signature='28s',
    # 0x3a   58
    notes=le.uint8.array(196),
    # 0xfe  254
)

# https://ist.uwaterloo.ca/~schepers/formats/GEOS.TXT
_INFO_BLOCK = le.Struct(
    # 0x02 / 0xfe - cvt leaves out the word-size pointer at the start
    #        so sectors are only 254 bytes long
    # > Information sector ID bytes (03 15 BF). The "03" is  likely
    # > the bitmap width, and the "15" is likely the bitmap height,
    # > but rare exceptions do exist to this!
    id_bytes='3s', # > 03 15 bf
    # 0x101
    # > Icon bitmap (sprite format, 63 bytes)
    icon=le.uint8.array(63),
    # 0x140
    # > C64 filetype (same as that from the directory entry)
    filetype='uint8',
    # > GEOS filetype (same as that from the directory entry)
    geos_filetype='uint8',
    # > GEOS file structure (same as that from the dir entry)
    geos_structure='uint8',
    # 0x143
    # > Program load address
    load_address='uint16',
    # > Program end address (only with accessories)
    end_address='uint16',
    # > Program start address
    start_address='uint16',
    # 0x149
    # > Class text (terminated with a $00)
    class_text='20s',

    # 0x61 / 0x15d
    # > Author (with application data: name  of  application  disk,
    # > terminated with a $00. This string may not  necessarily  be
    # > set, or it may contain invalid data)
    #author='20s',
    # 0x75 / 0x171
    # > 75-88: If a document, the name of the application that created it.
    ##application='20s',
    # 0x89/0x185
    # > 89-9F: Available for applications, unreserved.
    #unreserved='23s',

    # here the font INFO section diverges from the standard one
    # https://www.lyonlabs.org/commodore/onrequest/geos/geos-fonts.html
    # 0x61 / 0x15d
    O_GHSETLEN=le.uint16.array(15),
    skip='uint8',
    # 0x80 / 0x17c
    O_GHFONTID='uint16',
    # 0x82 / 0x17e
    O_GHPTSIZES=le.uint16.array(15),
    # 0xa0 / 0x19c

    # > A0-FF: Description (terminated with a $00)
    description='96s',
)

# font_id values
# http://www.zimmers.net/geos/docs/fontfile.txt
# 0    BSW            13   Tilden
# 1    University     14   Evans
# 2    California     15   Durant
# 3    Roma           16   Telegraph
# 4    Dwinelle       17   Superb
# 5    Cory           18   Bowditch
# 6    Tolman         19   Ormond
# 7    Bubble         20   Elmwood
# 8    Fontknox       21   Hearst
# 9    Harmon         21   Brennens (BUG)
# 10   Mykonos        23   Channing
# 11   Boalt          24   Putnam
# 12   Stadium        25   LeConte


# >   If the file is a VLIR, then the RECORD block is of interest. This  single
# > sector is made up of up to 127 track/sector pointers, each of  which  point
# > to program sections (called RECORDS). VLIR files are comprised of  loadable
# > RECORDS (overlays, if you wish to use PC terminology). The first RECORD  is
# > what is always loaded first when you run that application. After that,  the
# > OS loads whatever RECORD it needs.
# >   When a T/S link of $00/$00 is encountered, we  are  at  the  end  of  the
# > RECORD block. If the T/S  link  is  a  $00/$FF,  then  the  record  is  not
# > available.

# https://ist.uwaterloo.ca/~schepers/formats/CVT.TXT
# >   Note that the RECORD block is modified  from  the  original  GEOS  entry.
# > Instead of containing the track and sector  references,  we  now  have  the
# > sector count and the size of the last sector in the chain
_RECORD_ENTRY = le.Struct(
    sector_count='uint8',
    last_size='uint8',
)
_RECORD_BLOCK = _RECORD_ENTRY.array(127)


def _load_geos_cvt(instream, merge_mega):
    """Load a GEOS ConVerT container."""
    dir_entry = _DIR_BLOCK.read_from(instream)
    sig_block = _SIG_BLOCK.read_from(instream)
    logging.debug(
        'filetype: %s',
        _C64_FILETYPES.get(dir_entry.filetype, dir_entry.filetype)
    )
    logging.debug(
        'GEOS structure: %s',
        _GEOS_STRUCTURES.get(dir_entry.geos_structure, dir_entry.geos_structure)
    )
    logging.debug(
        'GEOS filetype: %s',
        _GEOS_FILETYPES.get(dir_entry.geos_filetype, dir_entry.geos_filetype)
    )
    logging.debug(
        'timestamp: %02d-%02d-%02d %02d:%02d',
        dir_entry.year, dir_entry.month, dir_entry.day,
        dir_entry.hour, dir_entry.minute
    )
    logging.debug('signature: %s', bytes(sig_block.signature).decode('ascii', 'replace'))
    info_block = _INFO_BLOCK.read_from(instream)
    logging.debug('class: %s', info_block.class_text.decode('ascii', 'replace'))
    icon = Raster.from_bytes(tuple(info_block.icon), width=24)
    record_block = _RECORD_BLOCK.read_from(instream)
    logging.debug(info_block)
    logging.debug(list(x for x in record_block if not bytes(x) == b'\0\xff'))
    if dir_entry.geos_filetype != _GEOS_FONT_TYPE:
        logging.warning(
            'Not a GEOS font file: incorrect filetype %x.',
            dir_entry.geos_filetype
        )
    comment = '\n\n'.join((
        info_block.description.decode('ascii', 'replace').replace('\r', '\n'),
        icon.as_text(),
    ))
    fonts = []
    for data_size, ghptsize in zip(
            info_block.O_GHSETLEN,
            info_block.O_GHPTSIZES
        ):
        if not ghptsize or not data_size:
            continue
        # ptsize is (font_id << 6) + point_size
        font_id, height = divmod(ghptsize, 1<<6)
        logging.debug('Loading font id %d height %d', font_id, height)
        anchor = instream.tell()
        try:
            font = _load_geos_vlir(instream)
        except ValueError as e:
            logging.warning(
                'Could not load font id %d size %d: %s', font_id, height, e
            )
            # flag to skip to next sector
            font = None
        true_data_size = instream.tell() - anchor
        if true_data_size != data_size:
            logging.warning(
                'Actual size 0x%x differs from reported size 0x%x',
                true_data_size, data_size
            )
        # go to end of sector
        # 254 bytes per sector - the cvt does not store the initial pointer
        nxt = ceildiv(true_data_size, 254) * 254
        instream.seek(anchor + nxt)
        if font is not None:
            font = font.modify(
                family=dir_entry.filename.rstrip(b'\xa0').decode('ascii', 'replace'),
                comment=comment,
                font_id=font_id,
            )
            fonts.append(font)
    # mega fonts: glyphs are divided over multiple strikes
    # undefined glyphs are given as 1-pixel-wide
    # last strike contains only empty or blank glyphs
    last_empty = all(_g.is_blank() for _g in fonts[-1].glyphs)
    # GHPTSIZE differs between strikes but actual pixel-size is the same
    # sometimes the last (empty) strike has a different height
    id_sizes = set((_f.font_id, _f.pixel_size) for _f in fonts[:-1])
    # if all strikes have the same id and pixel_size, assume this is a mega font
    if merge_mega and len(id_sizes) == 1 and last_empty:
        logging.info('Mega font detected, merging.')
        # take glyph of maximum width from each strike
        selected = (
            max(glyphs, key=lambda _g: _g.width)
            for glyphs in zip(*(_f.glyphs for _f in fonts))
        )
        fonts = (fonts[0].modify(glyphs=selected),)
    return fonts



"""
monobit.formats.dec - DEC Dynamically Redefined Character Set

(c) 2022--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# DRCS format documentation
# https://vt100.net/dec/vt320/soft_characters
# https://vt100.net/docs/vt510-rm/DECDLD


import shlex
import logging

from ..storage import loaders, savers
from ..streams import FileFormatError
from ..font import Font
from ..glyph import Glyph
from ..binary import ceildiv
from ..properties import reverse_dict

@loaders.register(
    magic=(b'\x90', b'\x1bP'),
    name='dec'
)
def load_dec_drcs(instream, where=None):
    """Load character-cell fonts from DEC DRCS file."""
    dec_glyphs, dec_props = _read_drcs(instream)
    props, count, first_codepoint = _parse_drcs_props(dec_props)
    glyphs = _parse_drcs_glyphs(dec_glyphs, first_codepoint)
    if len(glyphs) != count:
        logging.warning('Expected %d glyphs, found %d.', count, len(glyphs))
    return Font(glyphs, **props)

@savers.register(linked=load_dec_drcs)
def save_dec_drcs(fonts, outstream, where=None, *, use_8bit:bool=False):
    """Write font to a DEC DRCS file."""
    if len(fonts) > 1:
        raise FileFormatError('Can only save one font to dec-drcs file.')
        # check if font is fixed-width and fixed-height
    font, = fonts
    if font.spacing != 'character-cell':
        raise FileFormatError(
            'This format only supports character-cell fonts.'
        )
    # ensure codepoint values are set if possible
    font = font.label(codepoint_from=font.encoding)
    # upper size limits vary by device, not enforced.
    # lower sizes would conflict with vt200 size values
    if font.raster_size.x < 5 or font.raster_size.y < 1:
        raise FileFormatError(
            'This format only supports fonts of 5px or wider and 1px or taller.'
        )
    _write_dec_drcs(font, outstream)


##########################################################################

# initial and final escape sequences
_ESC_START = (b'\x90', b'\x1bP')
_ESC_END = (b'\x9c', b'\x1b\\')

# parameters
_DEC_PARMS = (
    'Pfn', # Font number
    # Selects the DRCS font buffer to load.
    # The VT320 has one DRCS font buffer. Pfn has two valid values, 0 and
    # 1. Both values refer to the same DRCS buffer.

    'Pcn', # Starting character
    # Selects where to load the first character in the DRCS font buffer.
    # The location corresponds to a location in the ASCII code table.
    # Pcn is affected by the character set size. (See Pcss below.) In a
    # 94-character set, a Pcn value of 0 or 1 means that the first soft
    # character is loaded into position 2/1 of the character table. In a
    # 96-character set, a Pcn value of 0 means the first character is
    # loaded into position 2/0 of the character table. The greatest Pcn
    # value is 95 (position 7/15).

    'Pe', # Erase control
    # Selects which characters to erase from the DRCS buffer before
    # loading the new font.
    # 0 = erase all characters in the DRCS buffer with this number,
    #     width and rendition.
    # 1 = erase only characters in locations being reloaded.
    # 2 = erase all renditions of the soft character set (normal, bold,
    #     80-column, 132-column).

    'Pcmw', # Character matrix width
    # Selects the maximum character cell width. VT300 modes:
    #     0 = 15 pixels wide for 80 columns,
    #         9 pixels wide for 132 columns. (Default)
    #     1 = illegal.
    #     2 = 5 × 10 pixel cell
    #     3 = 6 × 10 pixel cell
    #     4 = 7 × 10 pixel cell
    #     5 = 5 pixels wide.
    #     6 = 6 pixels wide.
    #     ...
    #     15 = 15 pixels wide.
    # If you omit a Pcmw value, the terminal uses the default character
    # width. Any Pcmw value over 15 is illegal.
    # Use Pcmw values 2 through 4 with VT220 compatible software. Remember
    # that VT220 fonts appear different VT320. Fonts designed specifically
    # for the VT320 should use values 5 through 15.

    'Pss', # Font set size (also 'Pw')
    # Defines the screen width and screen height for this font.
    # 0,1 = 80 columns, 24 lines. (default)
    #   2 = 132 columns, 24 lines
    #  11 = 80 columns, 36 lines
    #  12 = 132 columns, 36 lines
    #  21 = 80 columns, 48 lines
    #  22 = 132 columns, 48 lines

    'Pt', # Text or full-cell
    # Defines the font as a text font or full-cell font.
    #    0 = text. (Default)
    #    1 = text.
    #    2 = full cell.
    # Full-cell fonts can individually address all pixels in a cell.
    # Text fonts cannot individually address all pixels. If you specify a
    # text cell, the terminal automatically performs spacing and centering
    # of the characters.

    'Pcmh', # Character matrix height
    # Selects the maximum character cell height.
    #    0 = 12 pixels high. (Default)
    #    1 = 1 pixel high.
    #    2 = 2 pixels high.
    #    3 = 3 pixels high.
    #    ...
    #    12 = 12 pixels high.
    # Pcmh values over 12 are illegal. If the value of Pcmw is 2, 3 or 4,
    # Pcmh is ignored.

    'Pcss', # Character set size
    # Defines the character set as a 94- or 96-character graphic set.
    #    0 = 94-character set. (Default)
    #    1 = 96-character set.
    # The value of Pcss changes the meaning of the Pcn (starting
    # character) parameter above.
    # If Pcss = 0 (94-character set)
    #    The terminal ignores any attempt to load characters into the 2/0
    #    or 7/15 table positions.
    #    1  column 2/row 1
    #    ...
    #    94 column 7/row 14
    # If Pcss = 1 (96-character set)
    #    0 column 2/row 0
    #    ...
    #    95 column 7/row 15

    'Dscs', # Character Set Name
    # Dscs defines the character set name. It consists of from one to
    # three characters. The last character of the name is a character in
    # the range ‘0’ to ‘~’ (3016 to 7E16). There can be from zero to two
    # name characters preceding this one, in the range SP to ‘/’ (2016 to
    # 2F16). This name will be used in the Select Character Set (SCS)
    # sequence.
)

# terminal dimensions
_PSS_DIMS = {
    0: (80, 24),
    1: (80, 24),
    2: (132, 24),
    11: (80, 36),
    12: (132, 36),
    21: (80, 48),
    22: (132, 48),
}
_DEVICE_PATTERN = '{}x{} terminal'


# format-specific properties
_ERASE_CONTROL = {
    0: 'erase=font',
    1: 'erase=glyphs',
    2: 'erase=all',
}
_FONT_TYPE = {
    0: 'type=text',
    1: 'type=text',
    2: 'type=full-cell',
}

_JOINER = '='
_PFN_NAME = 'buffer'
_DSCS_NAME = 'dscs-id'
# recommmended default id for user-defined sets
# https://vt100.net/docs/vt510-rm/DECDLD
_DEFAULT_DSCS = ' @'

##########################################################################

def _read_char(f):
    c = f.read(1)
    while c in (b'\r', b'\n'):
        c = f.read(1)
    return c

def _read_args(f, sep, term=None):
    val, c = b'', b'*'
    while c and c != term:
        c = _read_char(f)
        if (c == term or not c) and not val:
            break
        if c in (sep, term):
            yield val
            val = b''
        else:
            val += c

def _read_dscs_name(f):
    dscs = []
    sep = _read_char(f)
    c = b''
    if sep != b'{':
        # { may have been absorbed by the Pcss read if there was no closing ;
        #raise FileFormatError(f'invalid Dscs separator {sep}')
        c = sep
    # read Dscs
    for _ in range(3):
        dscs.append(c)
        if c and ord(c) in range(0x30, 0x80):
            break
        if c and ord(c) not in range(0x20, 0x30):
            raise FileFormatError('invalid Dscs sequence')
        c = _read_char(f)
    return b''.join(dscs)

def _read_drcs(f):
    """Read a DEC DRCS file."""
    dcs = f.read(1)
    esc = dcs == b'\x1b'
    # one-byte x90 or two-byte esc P
    if esc:
        dcs += f.read(1)
    if not dcs in _ESC_START:
        raise FileFormatError('not a Dec DRCS file')
    argreader = _read_args(f, b';', b'{')
    dec_props = dict(zip(_DEC_PARMS[:-1], argreader))
    dec_props[_DEC_PARMS[-1]] = _read_dscs_name(f)
    term = _ESC_END[esc]
    # really shld be ESC \ but we only check 1 char
    term = term[0]
    glyphdefs = _read_args(f, b';', term)
    glyphdefs = list(glyphdefs)
    return glyphdefs, dec_props


def _convert_drcs_glyph(glyphdef):
    """Convert DRCS glyph to monobit glyph."""
    glyphbytes = (
        tuple(_b - ord(b'?') for _b in _block)
        for _block in glyphdef.split(b'/')
    )
    glyphbytes = zip(*glyphbytes)
    glyphstrs = tuple(
        ''.join(f'{_b:06b}' for _b in _pair[::-1])
        for _pair in glyphbytes
    )
    glyph = Glyph(glyphstrs, _0='0', _1='1')
    return glyph.turn(anti=1)


def _parse_drcs_glyphs(glyphdefs, first_codepoint):
    """Convert DRCS glyphs to monobit glyphs."""
    glyphs = (
        _convert_drcs_glyph(_g)
        for _g in glyphdefs
    )
    glyphs = tuple(
        _g.modify(codepoint=_cp)
        for _cp, _g in enumerate(glyphs, first_codepoint)
    )
    return glyphs

def _parse_drcs_props(dec_props):
    """Convert DRCS properties to yaff properties."""
    # determine glyph count from Pcss
    count = 96 if int(dec_props['Pcss']) else 94
    # determine starting codepoint from Pcn
    pcn = int(dec_props['Pcn'])
    first_codepoint = pcn + 0x20
    if count == 94 and not pcn:
        first_codepoint = 0x21
    # determine glyph width from Pcmw and Pss
    try:
        cols, rows = _PSS_DIMS[int(dec_props['Pss'])]
    except KeyError:
        raise FileFormatError(f"unknown value {dec_props['Pss']} for Pss")
    device = _DEVICE_PATTERN.format(cols, rows)
    pcmw = int(dec_props['Pcmw'])
    width, height = 0, 0
    if not pcmw:
        width = 9 if cols == 132 else 15
    elif pcmw >= 5:
        width = pcmw
    elif 4 >= pcmw >= 2:
        width = 3 + pcmw
        height = 10
    else:
        logging.warning('illegal value %d for Pcmw', pcmw)
    # determine glyph height from Pcmh
    if not height:
        height = int(dec_props['Pcmh']) or 12
    props = dict(
        encoding='ascii',
        raster_size=(width, height),
        device=device,
    )
    scs_id = dec_props['Dscs'].decode('ascii')
    # preserve unparsed properties
    props['dec_drcs'] = shlex.join((
        _JOINER.join((_DSCS_NAME, scs_id)),
        _JOINER.join((_PFN_NAME, str(int(dec_props['Pfn'])))),
        _ERASE_CONTROL[int(dec_props['Pe'])],
        _FONT_TYPE[int(dec_props['Pt'])],
    ))
    return props, count, first_codepoint

##########################################################################

def _write_dec_drcs(font, outstream, use_8bit=False):
    """Write a font to a DRCS file."""
    esc = not use_8bit
    # we can onl store the printable ascii range
    ascii = tuple(chr(_b) for _b in range(0x20, 0x80))
    glyphs = tuple(
        font.get_glyph(char=_c, missing='empty')
        for _c in ascii
    )
    # write 96 glyphs?
    is_big = not glyphs[0].is_blank() or not glyphs[-1].is_blank()
    if not is_big:
        glyphs = glyphs[1:-1]
    dec_props = _convert_to_drcs_props(font, is_big)
    outstream.write(_ESC_START[esc])
    outstream.write(b';'.join(
        str(dec_props[_k]).encode('ascii')
        for _k in _DEC_PARMS[:-1])
    )
    dscs = dec_props[_DEC_PARMS[-1]].encode('ascii')
    outstream.write(b'{')
    outstream.write(dscs)
    outstream.write(b'\n')
    for _g in glyphs:
        outstream.write(_convert_to_drcs_glyph(_g))
    outstream.write(_ESC_END[esc])
    # select character set (SCS) - no 8-bit variant?
    outstream.write(b' \x1b(%s\n' % (dscs,))

def _convert_to_drcs_props(font, is_big):
    # device spec
    pss_to_dims = reverse_dict(_PSS_DIMS)
    devices = {
        _DEVICE_PATTERN.format(*_k): _v
        for _k, _v in pss_to_dims.items()
    }
    pss = devices.get(font.device, 0)
    # format-specific
    pfn, pe, pt = 0, 1, 2
    dscs_id = _DEFAULT_DSCS
    try:
        unparsed = font.dec_drcs
    except AttributeError:
        pass
    else:
        uprops = shlex.split(unparsed)
        erase = reverse_dict(_ERASE_CONTROL)
        ftype = reverse_dict(_FONT_TYPE)
        for uprop in uprops:
            pe = erase.get(uprop, pe)
            pt = ftype.get(uprop, pt)
            key, _, value = uprop.split(_JOINER)
            if key == _PFN_NAME:
                pfn = value
            if key == _DSCS_NAME:
                dscs_id = value
    dec_props = {
        'Pfn': pfn,
        'Pcn': int(not is_big),  # we only define full sets
        'Pe': pe,
        'Pcmw': font.raster_size.x,
        'Pss': pss,
        'Pt': pt,
        'Pcmh': font.raster_size.y,
        'Pcss': int(is_big),
        'Dscs': dscs_id,
    }
    return dec_props

def _convert_to_drcs_glyph(glyph):
    # cut to sixel blocks
    nblocks = ceildiv(glyph.height, 6)
    blocks = (
        glyph.crop(
            top=_i*6,
            bottom=max(0, glyph.height-(_i+1)*6)
        )
        for _i in range(nblocks)
    )
    blocks = (_b.turn(clockwise=1) for _b in blocks)
    blockbytes = (_b.as_bytes(align='r') for _b in blocks)
    glyphdef = b'/'.join(
        bytes(_c + ord('?') for _c in _b)
        for _b in blockbytes
    ) + b';\n'
    return glyphdef



"""
monobit - tools for working with monochrome bitmap fonts

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# we need at least Python 3.7
import sys as _sys
assert _sys.version_info >= (3, 7)

from .constants import VERSION as __version__
from .pack import Pack, operations as _pack_operations
from .font import Font, operations as _operations
from .glyph import Glyph
from . import formats
from .storage import open_location, load, save, loaders, savers
from .encoding import charmaps, encoder
from .taggers import tagmaps
from .renderer import render
from .labels import Char, Codepoint, Tag


_operations.update(_pack_operations)
# inject font operations into main module namespace
globals().update(_operations)

# make dash-versions of operations available through dict
operations = {
    _name.replace('_', '-'): _func
    for _name, _func in _operations.items()
}



"""
monobit.containers - file containers

(c) 2021--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import io
import os
import sys
import time
import logging
import itertools
from contextlib import contextmanager
import zipfile
import tarfile
import gzip
import lzma
import bz2
from pathlib import Path, PurePath, PurePosixPath

from .streams import (
    MagicRegistry, FileFormatError,
    StreamBase, StreamWrapper, Stream, KeepOpen,
    get_suffix, open_stream, get_name
)

DEFAULT_ROOT = 'fonts'


def get_bytesio(bytestring):
    """Workaround as our streams objects require a buffer."""
    return io.BufferedReader(io.BytesIO(bytestring))


class ContainerFormatError(FileFormatError):
    """Incorrect container format."""


containers = MagicRegistry()

def open_container(file, mode, overwrite=False):
    """Open container of the appropriate type."""
    if isinstance(file, Container):
        return file
    if not file:
        # no-container, will throw errors when used
        return Container(None)
    container_type = _identify_container(file, mode, overwrite)
    container = container_type(file, mode, overwrite=overwrite)
    logging.debug("Opening %s container `%s` for '%s'.", container_type.__name__, container.name, mode)
    return container

def _identify_container(file, mode, overwrite):
    """Get container of the appropriate type."""
    if not file:
        raise ValueError('No location provided.')
    # if it already is a directory there is no choice
    if isinstance(file, (str, Path)) and Path(file).is_dir():
        container_type = DirContainer
    else:
        container_type = containers.identify(file, do_open=(mode == 'r'))
    if not container_type:
        suffix = get_suffix(file)
        # output to file with no suffix - default to directory
        if mode == 'w' and not suffix and isinstance(file, (str, Path)):
            return DirContainer
        # no container type found
        raise ContainerFormatError('Expected container format, got non-container stream.')
    return container_type


class Container(StreamBase):
    """Base class for container types."""

    def __iter__(self):
        """List contents."""
        raise NotImplementedError

    def open(self, name, mode):
        """Open a binary stream in the container."""
        raise NotImplementedError

    def unused_name(self, stem, suffix):
        """Generate unique name for container file."""
        for i in itertools.count():
            if i:
                filename = '{}.{}.{}'.format(stem, i, suffix)
            else:
                filename = '{}.{}'.format(stem, suffix)
            if filename not in self:
                return filename

###################################################################################################
# directory

class DirContainer(Container):
    """Treat directory tree as a container."""

    def __init__(self, path, mode='r', *, overwrite=False):
        """Create directory wrapper."""
        # if empty path, this refers to the whole filesystem
        if not path:
            path = ''
        self._path = Path(path)
        # mode really should just be 'r' or 'w'
        mode = mode[:1]
        if mode == 'w':
            logging.debug('Creating directory `%s`', self._path)
            # exist_ok raises FileExistsError only if the *target* already exists, not the parents
            self._path.mkdir(parents=True, exist_ok=overwrite)
        super().__init__(None, mode, str(self._path))

    def open(self, name, mode):
        """Open a stream in the container."""
        # mode in 'r', 'w'
        mode = mode[:1]
        pathname = Path(name)
        if mode == 'w':
            path = pathname.parent
            logging.debug('Creating directory `%s`', self._path / path)
            (self._path / path).mkdir(parents=True, exist_ok=True)
        # provide name relative to directory container
        file = Stream(self._path / pathname, mode=mode, name=str(pathname), overwrite=True)
        return file

    def __iter__(self):
        """List contents."""
        # don't walk the whole filesystem - no path is no contents
        if not self._path:
            return ()
        return (
            str((Path(_r) / _f).relative_to(self._path))
            for _r, _, _files in os.walk(self._path)
            for _f in _files
        )

    def __contains__(self, name):
        """File exists in container."""
        return (self._path / name).exists()


###################################################################################################
# zip archive

@containers.register('.zip', magic=(b'PK\x03\x04',))
class ZipContainer(Container):
    """Zip-file wrapper."""

    def __init__(self, file, mode='r', *, overwrite=False):
        """Create wrapper."""
        # mode really should just be 'r' or 'w'
        mode = mode[:1]
        # reading zipfile needs a seekable stream, drain to buffer if needed
        # note you can only do this once on the input stream!
        if (mode == 'r' and not isinstance(file, (str, Path)) and not file.seekable()):
            # note file is externally provided so we shouldn't close it
            # but the BytesIO is ours
            stream = get_bytesio(file.read())
        else:
            stream = open_stream(file, mode, overwrite=overwrite)
        # create the zipfile
        try:
            self._zip = zipfile.ZipFile(stream, mode, compression=zipfile.ZIP_DEFLATED)
        except zipfile.BadZipFile as exc:
            raise ContainerFormatError(exc) from exc
        super().__init__(stream, mode, self._zip.filename)
        # on output, put all files in a directory with the same name as the archive (without suffix)
        if mode == 'w':
            self._root = Path(self.name).stem or DEFAULT_ROOT
        else:
            self._root = ''
        # output files, to be written on close
        self._files = []


    def close(self):
        """Close the zip file, ignoring errors."""
        if self.mode == 'w' and not self.closed:
            for file in self._files:
                logging.debug('Writing out `%s` to zip container `%s`.', file.name, self.name)
                bytearray = file.getvalue()
                file.close()
                self._zip.writestr(file.name, bytearray)
        try:
            self._zip.close()
        except EnvironmentError:
            # e.g. BrokenPipeError
            pass
        super().close()
        self.closed = True

    def __iter__(self):
        """List contents."""
        return (
            str(PurePosixPath(_name).relative_to(self._root))
            for _name in self._zip.namelist()
        )

    def open(self, name, mode):
        """Open a stream in the container."""
        # using posixpath for internal paths in the archive
        # as forward slash should always work, but backslash would fail on unix
        filename = str(PurePosixPath(self._root) / name)
        mode = mode[:1]
        # always open as binary
        logging.debug('Opening file `%s` on zip container `%s`.', filename, self.name)
        if mode == 'r':
            return Stream(self._zip.open(filename, mode), mode=mode)
        else:
            # stop BytesIO from being closed until we want it to be
            newfile = KeepOpen(io.BytesIO(), mode=mode, name=filename)
            if filename in self._files:
                logging.warning('Creating multiple files of the same name `%s`.', filename)
            self._files.append(newfile)
            return newfile



###################################################################################################
# tar archive

@containers.register('.tar')
class TarContainer(Container):
    """Tar-file wrapper."""

    def __init__(self, file, mode='r',*, overwrite=False):
        """Create wrapper."""
        # mode really should just be 'r' or 'w'
        mode = mode[:1]
        # reading tarfile needs a seekable stream, drain to buffer if needed
        # note you can only do this once on the input stream!
        if (mode == 'r' and not isinstance(file, (str, Path)) and not file.seekable()):
            # note file is externally provided so we shouldn't close it
            # but the BytesIO is ours
            stream = get_bytesio(file.read())
        else:
            stream = open_stream(file, mode, overwrite=overwrite)
        # create the tarfile
        try:
            self._tarfile = tarfile.open(fileobj=stream, mode=mode)
        except tarfile.ReadError as exc:
            raise ContainerFormatError(exc) from exc
        super().__init__(stream, mode, self._tarfile.name)
        # on output, put all files in a directory with the same name as the archive (without suffix)
        if mode == 'w':
            self._root = Path(self.name).stem or DEFAULT_ROOT
        else:
            self._root = ''
        # output files, to be written on close
        self._files = []

    def close(self):
        """Close the tar file, ignoring errors."""
        if self.mode == 'w' and not self.closed:
            for file in self._files:
                name = file.name
                logging.debug('Writing out `%s` to tar container `%s`.', name, self.name)
                tinfo = tarfile.TarInfo(name)
                tinfo.mtime = time.time()
                tinfo.size = len(file.getvalue())
                file.seek(0)
                self._tarfile.addfile(tinfo, file)
                file.close()
        try:
            self._tarfile.close()
        except EnvironmentError:
            # e.g. BrokenPipeError
            pass
        super().close()
        self.closed = True

    def __iter__(self):
        """List contents."""
        # list regular files only, skip symlinks and dirs and block devices
        return (_ti.name for _ti in self._tarfile.getmembers() if _ti.isfile())

    def open(self, name, mode):
        """Open a stream in the container."""
        name = str(PurePosixPath(self._root) / name)
        mode = mode[:1]
        # always open as binary
        logging.debug('Opening file `%s` on tar container `%s`.', name, self.name)
        if mode == 'r':
            file = self._tarfile.extractfile(name)
            # .name is not writeable, so we need to wrap
            return Stream(file, mode, name=name)
        else:
            # stop BytesIO from being closed until we want it to be
            newfile = KeepOpen(io.BytesIO(), mode=mode, name=name)
            if name in self._files:
                logging.warning('Creating multiple files of the same name `%s`.', name)
            self._files.append(newfile)
            return newfile


###################################################################################################
# single-file compression

class Compressor(Container):
    """Base class for compression helpers."""

    format = ''
    compressor = None

    def __init__(self, infile, mode='r', *, overwrite=False):
        stream = Stream(infile, mode, overwrite=overwrite)
        super().__init__(stream, mode)
        # drop the .gz etc
        last_suffix = get_suffix(self.name)
        if last_suffix == self.format:
            self._content_name = self.name[:-1-len(last_suffix)]
        else:
            self._content_name = self.name

    def __iter__(self):
        return iter((self._content_name,))

    def open(self, name='', mode=''):
        """Open a stream in the container."""
        mode = mode[:1] or self.mode
        wrapped = self.compressor.open(self._stream, mode + 'b')
        wrapped = Stream(wrapped, mode, name=self._content_name)
        logging.debug(
            "Opening %s-compressed stream `%s` on `%s` for mode '%s'",
            self.format, wrapped.name, self.name, mode
        )
        return wrapped


@containers.register('.gz', magic=(b'\x1f\x8b',))
class GzipCompressor(Compressor):
    compressor = gzip

@containers.register('.xz', magic=(b'\xFD7zXZ\x00',))
class XZCompressor(Compressor):
    compressor = lzma

# the magic is a 'maybe'
@containers.register('.lzma', magic=(b'\x5d\0\0',))
class LzmaCompressor(Compressor):
    compressor = lzma

@containers.register('.bz2', magic=(b'BZh',))
class Bzip2Compressor(Compressor):
    compressor = bz2


##############################################################################
# Mac data/resource fork containers

class MacContainer(Container):
    """Mac data/resource fork container."""

    parse = None

    def __init__(self, file, mode='r', *, overwrite=False):
        """Create wrapper."""
        # mode really should just be 'r' or 'w'
        mode = mode[:1]
        if mode != 'r':
            raise ContainerFormatError(
                'Writing to Mac resource container is not implemented.'
            )
        with open_stream(file, mode, overwrite=overwrite) as stream:
            self._fork_name, self._data, self._rsrc = self.parse(stream)
        super().__init__(None, mode, stream.name)

    def __iter__(self):
        """List contents."""
        contents = []
        if self._data:
            contents.append(f'{self._fork_name}.data')
        if self._rsrc:
            contents.append(f'{self._fork_name}.rsrc')
        return iter(contents)

    def open(self, name, mode):
        """Open a stream in the container."""
        # using posixpath for internal paths in the archive
        # as forward slash should always work, but backslash would fail on unix
        mode = mode[:1]
        if mode != 'r':
            raise ContainerFormatError(
                'Writing to Mac resource container is not implemented.'
            )
        filename = str(name)[-4:].lower()
        if filename not in ('data', 'rsrc'):
            raise FileNotFoundError(
                'Stream name on Mac resource container must end with `rsrc` or `data`'
            )
        filename = f'{self._fork_name}.{filename}'
        # always open as binary
        logging.debug(
            'Opening fork `%s` on %s `%s`.', filename,
            type(self).__name__, self.name
        )
        if filename.endswith('data'):
            fork = self._data
        else:
            fork = self._rsrc
        if not fork:
            raise FileNotFoundError(filename)
        newfile = Stream(get_bytesio(fork), mode=mode, name=filename)
        return newfile


###############################################################################
# BinHex 4.0 container
# https://files.stairways.com/other/binhex-40-specs-info.txt

from binascii import crc_hqx
from itertools import zip_longest

from .struct import big_endian as be


_BINHEX_CODES = (
    '''!"#$%&'()*+,-012345689@ABCDEFGHIJKLMNPQRSTUVXYZ[`abcdefhijklmpqr'''
)
_BINHEX_CODEDICT = {_BINHEX_CODES[_i]: _i for _i in range(64)}

_BINHEX_HEADER = be.Struct(
    type='4s',
    auth='4s',
    flag='uint16',
    dlen='uint32',
    rlen='uint32',
)
_CRC = be.Struct(
    crc='uint16',
)

def _parse_binhex(stream):
    """Parse a BinHex 4.0 file."""
    front, binhex, *back = stream.text.read().split(':')
    if 'BinHex 4.0' not in front:
        logging.warning('No BinHex 4.0 signature found.')
    back = ''.join(back).strip()
    if back:
        logging.warning('Additional data found after BinHex section: %r', back)
    binhex = ''.join(binhex.split('\n'))
    # decode into 6-bit ints
    data = (_BINHEX_CODEDICT[_c] for _c in binhex)
    # convert to bit sequence
    bits = ''.join(bin(_d)[2:].zfill(6) for _d in data)
    # group into chunks of 8
    args = [iter(bits)] * 8
    octets = (''.join(_t) for _t in zip_longest(*args, fillvalue='0'))
    # convert to bytes
    bytestr = bytes(int(_s, 2) for _s in octets)
    # find run-length encoding marker
    chunks = bytestr.split(b'\x90')
    out = bytearray(chunks[0])
    for c in chunks[1:]:
        if c:
            # run-length byte
            repeat = c[0]
        else:
            # ...\x90\x90... -> ...', '', '...
            repeat = 0x90
        if not repeat:
            # zero-byte is placeholder for just 0x90
            out += b'\x90'
        else:
            # apply RLE. the last byte counts as the first of the run
            out += out[-1:] * (repeat-1)
        out += c[1:]
    # decode header
    length = out[0]
    name = bytes(out[1:1+length]).decode('mac-roman')
    if out[1+length] != 0:
        logging.warning('No null byte after name')
    header = _BINHEX_HEADER.from_bytes(out, 2+length)
    logging.debug(header)
    offset = 2 + length + _BINHEX_HEADER.size
    crc_header = out[:offset]
    hc = _CRC.from_bytes(out, offset)
    offset += _CRC.size
    if crc_hqx(crc_header, 0) != hc.crc:
        logging.error('CRC fault in header')
    data = out[offset:offset+header.dlen]
    offset += header.dlen
    dc = _CRC.from_bytes(out, offset)
    offset += _CRC.size
    rsrc = out[offset:offset+header.rlen]
    offset += header.rlen
    rc = _CRC.from_bytes(out, offset)
    if crc_hqx(data, 0) != dc.crc:
        logging.error('CRC fault in data fork')
    if crc_hqx(rsrc, 0) != rc.crc:
        logging.error('CRC fault in resource fork')
    return name, data, rsrc


@containers.register(
    '.hqx', magic=(
        b'(This file must be converted',
        b'\r(This file must be converted',
    ),
)
class BinHexContainer(MacContainer):
    """BinHex 4.0 Container."""
    parse = staticmethod(_parse_binhex)


##############################################################################
# MacBinary container
# v1: https://www.cryer.co.uk/file-types/b/bin_/original_mac_binary_format_proposal.htm
# v2: https://files.stairways.com/other/macbinaryii-standard-info.txt
# v2 defines additional fields inside an area zeroed in v1. we can ignore them.

from .binary import align


_MACBINARY_HEADER = be.Struct(
    # Offset 000-Byte, old version number, must be kept at zero for compatibility
    old_version='byte',
    # Offset 001-Byte, Length of filename (must be in the range 1-63)
    filename_length='byte',
    # Offset 002-1 to 63 chars, filename (only "length" bytes are significant).
    filename='63s',
    # Offset 065-Long Word, file type (normally expressed as four characters)
    file_type='4s',
    # Offset 069-Long Word, file creator (normally expressed as four characters)
    file_creator='4s',
    # Offset 073-Byte, original Finder flags
    original_finder_flags='byte',
    # Offset 074-Byte, zero fill, must be zero for compatibility
    zero_0='byte',
    # Offset 075-Word, file's vertical position within its window.
    window_vert='word',
    # Offset 077-Word, file's horizontal position within its window.
    window_horiz='word',
    # Offset 079-Word, file's window or folder ID.
    window_id='word',
    # Offset 081-Byte, "Protected" flag (in low order bit).
    protected='byte',
    # Offset 082-Byte, zero fill, must be zero for compatibility
    zero_1='byte',
    # Offset 083-Long Word, Data Fork length (bytes, zero if no Data Fork).
    data_length='dword',
    # Offset 087-Long Word, Resource Fork length (bytes, zero if no R.F.).
    rsrc_length='dword',
    # Offset 091-Long Word, File's creation date
    creation_date='dword',
    # Offset 095-Long Word, File's "last modified" date.
    last_modified_date='dword',
    # Offset 099-Word, length of Get Info comment to be sent after the resource
    # fork (if implemented, see below).
    get_info_length='word',
    # *Offset 101-Byte, Finder Flags, bits 0-7. (Bits 8-15 are already in byte 73)
    finder_flags='byte',
    # *Offset 116-Long Word, Length of total files when packed files are unpacked.
    packed_length='dword',
    # *Offset 120-Word, Length of a secondary header.  If this is non-zero,
    #              Skip this many bytes (rounded up to the next multiple of 128)
    #              This is for future expansion only, when sending files with
    #              MacBinary, this word should be zero.
    second_header_length='dword',
    # *Offset 122-Byte, Version number of Macbinary II that the uploading program
    # is written for (the version begins at 129)
    writer_version='byte',
    # *Offset 123-Byte, Minimum MacBinary II version needed to read this file
    # (start this value at 129 129)
    reader_version='byte',
    # *Offset 124-Word, CRC of previous 124 bytes
    crc='word',
    # from v1 desc:
    # > 126 2 Reserved for computer type and OS ID
    # > (this field will be zero for the current Macintosh).
    reserved='word',
    # *This is newly defined for MacBinary II.
)

def _parse_macbinary(stream):
    """Parse a MacBinary file."""
    data = stream.read()
    header = _MACBINARY_HEADER.from_bytes(data)
    ofs = 128
    if header.old_version != 0:
        raise FileFormatError(
            'Not a MacBinary file: incorrect version field'
            f' ({header.old_version}).'
        )
    if header.writer_version > 128:
        ofs += align(header.second_header_length, 7)
    data_fork = data[ofs:ofs+header.data_length]
    ofs += align(header.data_length, 7)
    rsrc_fork = data[ofs:ofs+header.rsrc_length]
    name = header.filename.decode('mac-roman').strip()
    return name, data_fork, rsrc_fork


@containers.register('.bin')
class MacBinaryContainer(MacContainer):
    """MacBinary Container."""
    parse = staticmethod(_parse_macbinary)


##############################################################################
# AppleSingle/AppleDouble container
# v1: see https://web.archive.org/web/20160304101440/http://kaiser-edv.de/documents/Applesingle_AppleDouble_v1.html
# v2: https://web.archive.org/web/20160303215152/http://kaiser-edv.de/documents/AppleSingle_AppleDouble.pdf
# the difference between v1 and v2 affects the file info sections
# not the resource fork which is what we care about

_APPLESINGLE_MAGIC = 0x00051600
_APPLEDOUBLE_MAGIC = 0x00051607


_APPLE_HEADER = be.Struct(
    magic='uint32',
    version='uint32',
    home_fs='16s',
    number_entities='uint16',
)
_APPLE_ENTRY = be.Struct(
    entry_id='uint32',
    offset='uint32',
    length='uint32',
)

# Entry IDs
_ID_DATA = 1
_ID_RESOURCE = 2
_ID_NAME = 3

_APPLE_ENTRY_TYPES = {
    1: 'data fork',
    2: 'resource fork',
    3: 'real name',
    4: 'comment',
    5: 'icon, b&w',
    6: 'icon, color',
    7: 'file info', # v1 only
    8: 'file dates info', # v2
    9: 'finder info',
    # the following are all v2
    10: 'macintosh file info',
    11: 'prodos file info',
    12: 'ms-dos file info',
    13: 'short name',
    14: 'afp file info',
    15: 'directory id',
}


def _parse_apple_container(stream):
    """Parse an AppleSingle or AppleDouble file."""
    data = stream.read()
    header = _APPLE_HEADER.from_bytes(data)
    if header.magic == _APPLESINGLE_MAGIC:
        container = 'AppleSingle'
    elif header.magic == _APPLEDOUBLE_MAGIC:
        container = 'AppleDouble'
    else:
        raise FileFormatError('Not an AppleSingle or AppleDouble file.')
    entry_array = _APPLE_ENTRY.array(header.number_entities)
    entries = entry_array.from_bytes(data, _APPLE_HEADER.size)
    name, data_fork, rsrc_fork = '', b'', b''
    for i, entry in enumerate(entries):
        entry_type = _APPLE_ENTRY_TYPES.get(entry.entry_id, 'unknown')
        logging.debug(
            '%s container: entry #%d, %s [%d]',
            container, i, entry_type, entry.entry_id
        )
        if entry.entry_id == _ID_RESOURCE:
            rsrc_fork = data[entry.offset:entry.offset+entry.length]
        if entry.entry_id == _ID_DATA:
            data_fork = data[entry.offset:entry.offset+entry.length]
        if entry.entry_id == _ID_NAME:
            name = data[entry.offset:entry.offset+entry.length]
            name = name.decode('mac-roman')
    return name, data_fork, rsrc_fork



@containers.register(
    '.as',
    magic=(
        _APPLESINGLE_MAGIC.to_bytes(4, 'big'),
    ),
)
class AppleSingleContainer(MacContainer):
    """AppleSingle Container."""
    parse = staticmethod(_parse_apple_container)


@containers.register(
    '.adf', #'rsrc',
    magic=(
        _APPLEDOUBLE_MAGIC.to_bytes(4, 'big'),
    ),
)
class AppleDoubleContainer(MacContainer):
    """AppleDouble Container."""
    parse = staticmethod(_parse_apple_container)



"""
monobit.formats.source - fonts embedded in C/Python/JS source files

(c) 2019--2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from io import BytesIO
import string

from ..binary import ceildiv
from ..storage import loaders, savers
from ..font import Font
from ..glyph import Glyph
from ..streams import FileFormatError
from ..basetypes import Coord
from .raw import load_bitmap


_C_PARAMS = dict(
    delimiters='{}',
    comment='//',
)

_JS_PARAMS = dict(
    delimiters='[]',
    comment='//',
)

_PY_PARAMS = dict(
    delimiters='[]',
    comment='#',
)

###################################################################################################

@loaders.register('c', 'cc', 'cpp', 'h', name='c')
def load_c(
        infile, where=None, *,
        identifier:str='',
        cell:Coord=(8, 8), count:int=-1, offset:int=0, padding:int=0,
        align:str='left', strike_count:int=1, strike_bytes:int=-1,
        first_codepoint:int=0
    ):
    """
    Extract font from bitmap encoded in C or C++ source code.

    identifier: text at start of line where bitmap starts. (default: first array literal {})
    cell: size X,Y of character cell (default: 8x8)
    offset: number of bytes in file before bitmap starts (default: 0)
    padding: number of bytes between encoded glyphs (default: 0)
    count: number of glyphs to extract (<=0 means all; default: all glyphs)
    align: alignment of glyph in byte (left for most-, right for least-significant; default: left)
    strike_count: number of glyphs in glyph row (<=0 for all; default: 1)
    strike_bytes: strike width in bytes (<=0 means as many as needed to fit the glyphs; default: as needed)
    first_codepoint: first code point in bitmap (default: 0)
    """
    return _load_coded_binary(
        infile, where, identifier=identifier,
        cell=cell, count=count, offset=offset, padding=padding,
        align=align, strike_count=strike_count, strike_bytes=strike_bytes,
        first_codepoint=first_codepoint,
        **_C_PARAMS
    )

@loaders.register('js', 'json', name='json')
def load_json(
        infile, where=None, *,
        identifier:str='',
        cell:Coord=(8, 8), count:int=-1, offset:int=0, padding:int=0,
        align:str='left', strike_count:int=1, strike_bytes:int=-1,
        first_codepoint:int=0
    ):
    """
    Extract font from bitmap encoded in JavaScript source code.

    identifier: text at start of line where bitmap starts (default: first list [])
    cell: size X,Y of character cell (default: 8x8)
    offset: number of bytes in file before bitmap starts (default: 0)
    padding: number of bytes between encoded glyphs (default: 0)
    count: number of glyphs to extract (<=0 means all; default: all)
    align: alignment of glyph in byte (left for most-, right for least-significant; default: left)
    strike_count: number of glyphs in glyph row (<=0 for all; default: 1)
    strike_bytes: strike width in bytes (<=0 means as many as needed to fit the glyphs; default: as needed)
    first_codepoint: first code point in bitmap (default: 0)
    """
    return _load_coded_binary(
        infile, where, identifier=identifier,
        cell=cell, count=count, offset=offset, padding=padding,
        align=align, strike_count=strike_count, strike_bytes=strike_bytes,
        first_codepoint=first_codepoint,
        **_JS_PARAMS
    )

@loaders.register('py', name='python')
def load_python(
        infile, where=None, *,
        identifier:str='',
        cell:Coord=(8, 8), count:int=-1, offset:int=0, padding:int=0,
        align:str='left', strike_count:int=1, strike_bytes:int=-1,
        first_codepoint:int=0
    ):
    """
    Extract font from bitmap encoded as a list in Python source code.

    identifier: text at start of line where bitmap starts (default: first list [])
    cell: size X,Y of character cell (default: 8x8)
    offset: number of bytes in file before bitmap starts (default: 0)
    padding: number of bytes between encoded glyphs (default: 0)
    count: number of glyphs to extract (<=0 means all; default: all)
    align: alignment of glyph in byte (left for most-, right for least-significant; default: left)
    strike_count: number of glyphs in glyph row (<=0 for all; default: 1)
    strike_bytes: strike width in bytes (<=0 means as many as needed to fit the glyphs; default: as needed)
    first_codepoint: first code point in bitmap (default: 0)
    """
    return _load_coded_binary(
        infile, where, identifier=identifier,
        cell=cell, count=count, offset=offset, padding=padding,
        align=align, strike_count=strike_count, strike_bytes=strike_bytes,
        first_codepoint=first_codepoint,
        **_PY_PARAMS
    )

@loaders.register(name='source')
def load_source(
        infile, where=None, *,
        identifier:str='', delimiters:str='{}', comment:str='//',
        cell:Coord=(8, 8), count:int=-1, strike_count:int=1, strike_bytes:int=-1,
        offset:int=0, padding:int=0, align:str='left',
        first_codepoint:int=0
    ):
    """
    Extract font from bitmap encoded in source code.

    identifier: text at start of line where bitmap starts (default: first delimiter)
    delimiters: pair of delimiters that enclose the bitmap (default: {})
    comment: string that introduces inline comment (default: //)
    cell: size X,Y of character cell (default: 8x8)
    offset: number of bytes in file before bitmap starts (default: 0)
    padding: number of bytes between encoded glyphs (default: 0)
    count: number of glyphs to extract (<=0 means all; default: all)
    align: alignment of glyph in byte (left for most-, right for least-significant; default: left)
    strike_count: number of glyphs in glyph row (<=0 for all; default: 1)
    strike_bytes: strike width in bytes (<=0 means as many as needed to fit the glyphs; default: as needed)
    first_codepoint: first code point in bitmap (default: 0)
    """

    return _load_coded_binary(
        infile, where, identifier=identifier,
        cell=cell, count=count, offset=offset, padding=padding,
        align=align, strike_count=strike_count, strike_bytes=strike_bytes,
        first_codepoint=first_codepoint,
        delimiters=delimiters, comment=comment
    )


def _load_coded_binary(
        infile, where, identifier, delimiters, comment,
        cell, count, offset, padding, align, strike_count, strike_bytes, first_codepoint
    ):
    """Load font from binary encoded in source code."""
    width, height = cell
    payload = _get_payload(infile.text, identifier, delimiters, comment)
    data = bytes(_int_from_c(_s) for _s in payload.split(',') if _s)
    bytesio = BytesIO(data[offset:])
    return load_bitmap(
        bytesio, width, height, count, padding, align, strike_count, strike_bytes, first_codepoint
    )

def _int_from_c(cvalue):
    """Parse integer from c code."""
    # suffixes
    while cvalue[-1:].lower() in ('u', 'l'):
        cvalue = cvalue[:-1]
    if cvalue.startswith('0') and cvalue[1:2] and cvalue[1:2] in string.digits:
        # C / Python-2 octal 0777
        cvalue = '0o' + cvalue[1:]
    # 0x, 0b, decimals - like Python
    return int(cvalue, 0)

def _get_payload(instream, identifier, delimiters, comment):
    """Find the identifier and get the part between delimiters."""
    start, end = delimiters
    for line in instream:
        if comment in line:
            line, _ = line.split(comment, 1)
        line = line.strip(' \r\n')
        if identifier in line:
            if identifier:
                _, line = line.split(identifier)
            if start in line:
                _, line = line.split(start)
                break
    else:
        raise ValueError('No payload with identifier `{}` found in file'.format(identifier))
    if end in line:
        line, _ = line.split(end, 1)
        return line
    payload = [line]
    for line in instream:
        if comment in line:
            line, _ = line.split(comment, 1)
        line = line.strip(' \r\n')
        if start in line:
            _, line = line.split(start, 1)
        if end in line:
            line, _ = line.split(end, 1)
            payload.append(line)
            break
        if line:
            payload.append(line)
    return ''.join(payload)


###################################################################################################

@savers.register('c', linked=load_c)
def save_c(fonts, outstream, where=None):
    """
    Save font to bitmap encoded in C source code.
    """
    return _save_coded_binary(fonts, outstream, 'char font_{compactname}[{bytesize}] = ', **_C_PARAMS)

@savers.register('py', 'python', linked=load_python)
def save_python(fonts, outstream, where=None):
    """
    Save font to bitmap encoded in Python source code.
    """
    return _save_coded_binary(fonts, outstream, 'font_{compactname} = ', **_PY_PARAMS)

@savers.register('json', linked=load_json)
def save_json(fonts, outstream, where=None):
    """
    Save font to bitmap encoded in JSON code.
    """
    return _save_coded_binary(fonts, outstream, '', **_JS_PARAMS)

@savers.register('source', linked=load_source)
def save_source(
        fonts, outstream, where=None, *,
        identifier:str, assign:str='=', delimiters:str='{}', comment:str='//',
    ):
    """
    Save font to bitmap encoded in source code.
    """
    return _save_coded_binary(fonts, outstream, f'{identifier} {assign} ', delimiters, comment)

def _save_coded_binary(fonts, outstream, assignment_pattern, delimiters, comment):
    """
    Generate bitmap encoded source code from a font.

    Args:
        fonts (List[Font]): Exactly one font must be given.
        outstream: Stream to write the source code to.
        assignment_pattern: Format pattern for the assignment statement. May include `compactname` amd `bytesize` variables.
        delimiters (str): Must contain two characters, building the opening and closing delimiters of the collection. E.g. []
        comment (str): Line Comment character(s). Currently not used.

    Raises:
        FileFormatError: If more the one Font is passed or if it is not a character-cell font.
        ValueError: If delimiter does not contain at least two characters.

    Returns:
        Font: Used font.
    """
    if len(fonts) > 1:
        raise FileFormatError('Can only save one font to source file.')
    font = fonts[0]
    outstream = outstream.text
    # check if font is fixed-width and fixed-height
    if font.spacing != 'character-cell':
        raise FileFormatError(
            'This format only supports character-cell fonts.'
        )
    if len(delimiters) < 2:
        raise ValueError('A start and end delimiter must be given. E.g. []')
    start_delimiter = delimiters[0]
    end_delimiter = delimiters[1]
    # build the identifier
    ascii_name = font.name.encode('ascii', 'ignore').decode('ascii')
    ascii_name = ''.join(_c if _c.isalnum() else '_' for _c in ascii_name)
    width, height = font.raster_size
    bytesize = ceildiv(width, 8) * height * len(font.glyphs)
    assignment = assignment_pattern.format(compactname=ascii_name, bytesize=bytesize)
    # emit code
    outstream.write(f'{assignment}{start_delimiter}\n')
    for glyph in font.glyphs:
        outstream.write('  ')
        for byte in glyph.as_bytes():
            outstream.write(f'0x{byte:02x}, ')
        outstream.write('\n')
    outstream.write(f'{end_delimiter}\n')
    return font
